# Copyright GlitterClient GmbH and GlitterClient contributors
# SPDX-License-Identifier: (Apache-2.0 AND CC-BY-4.0)
# Code is Apache-2.0 and docs are CC-BY-4.0

"""Set of utilities to support various functionalities of the driver.

Attributes:
    ops_map (dict): Mapping between operation strings and classes.
        E.g.: The string ``'CREATE'`` is mapped to
        :class:`~.CreateOperation`.
"""
from urllib.parse import urlparse, urlunparse

DEFAULT_NODE = 'http://sg1.testnet.glitter.link:26659'
DEFAULT_NODES = [
                 'http://sg1.testnet.glitter.link:26659', 'http://sg2.testnet.glitter.link:26659',
                 'http://sg3.testnet.glitter.link:26659', 'http://sg4.testnet.glitter.link:26659',
                 'http://sg7.testnet.glitter.link:26659', 'http://sg8.testnet.glitter.link:26659',
                 'http://sg9.testnet.glitter.link:26659', 'http://sg10.testnet.glitter.link:26659',
                 'http://sg11.testnet.glitter.link:26659', 'http://sg12.testnet.glitter.link:26659',
                 'http://sg13.testnet.glitter.link:26659', 'http://sg14.testnet.glitter.link:26659',
                 'http://sg15.testnet.glitter.link:26659',
                 ]


class CreateOperation:
    """Class representing the ``'CREATE'`` transaction operation."""


class TransferOperation:
    """Class representing the ``'TRANSFER'`` transaction operation."""


ops_map = {
    'CREATE': CreateOperation,
    'TRANSFER': TransferOperation,
}


def _normalize_operation(operation):
    """Normalizes the given operation string. For now, this simply means
    converting the given string to uppercase, looking it up in
    :attr:`~.ops_map`, and returning the corresponding class if
    present.

    Args:
        operation (str): The operation string to convert.

    Returns:
        The class corresponding to the given string,
        :class:`~.CreateOperation` or :class:`~TransferOperation`.

        .. important:: If the :meth:`str.upper` step, or the
            :attr:`~.ops_map` lookup fails, the given ``operation``
            argument is returned.

    """
    try:
        operation = operation.upper()
    except AttributeError:
        pass

    try:
        operation = ops_map[operation]()
    except KeyError:
        pass

    return operation


def _get_default_port(scheme):
    return 443 if scheme == 'https' else 9984


def normalize_url(node):
    """Normalizes the given node url"""
    if not node:
        node = DEFAULT_NODE
    elif '://' not in node:
        node = '//{}'.format(node)
    parts = urlparse(node, scheme='http', allow_fragments=False)
    port = parts.port if parts.port else _get_default_port(parts.scheme)
    netloc = '{}:{}'.format(parts.hostname, port)
    return urlunparse((parts.scheme, netloc, parts.path, '', '', ''))


def normalize_node(node, headers=None):
    """Normalizes given node as str or dict with headers"""
    headers = {} if headers is None else headers
    if isinstance(node, str):
        url = normalize_url(node)
        return {'endpoint': url, 'headers': headers}

    url = normalize_url(node['endpoint'])
    node_headers = node.get('headers', {})
    return {'endpoint': url, 'headers': {**headers, **node_headers}}


def normalize_nodes(*nodes, headers=None):
    """Normalizes given dict or array of driver nodes"""
    if not nodes:
        nodes = DEFAULT_NODES

    normalized_nodes = ()
    for node in nodes:
        normalized_nodes += (normalize_node(node, headers),)
    return normalized_nodes



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import datetime
import sphinx_rtd_theme

cwd = os.getcwd()
project_root = os.path.dirname(cwd)
sys.path.insert(0, project_root)

# -- Project information -----------------------------------------------------
import glitter_sdk
project = 'glitter_sdk'
author = 'glitter'
now = datetime.datetime.now()
copyright = str(now.year) + ', glitter'
version = glitter_sdk.__version__
release = glitter_sdk.__release__

# -- General configuration ---------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinxcontrib.httpdomain',
    'IPython.sphinxext.ipython_console_highlighting',
    'IPython.sphinxext.ipython_directive',
    'sphinx_tabs.tabs',
]
# -- Options for HTML output ----------------------------------------------
# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
exclude_patterns = []

html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_static_path = ['_static']
autodoc_member_order = 'bysource'
autodoc_default_options = {
    'show-inheritance': None,
}
exclude_patterns = ['_build']
pygments_style = 'sphinx'
todo_include_todos = True
suppress_warnings = ['image.nonlocal_uri']

htmlhelp_basename = 'glitter_python_driverdoc'

latex_elements = {}



# -- Options for manual page output ---------------------------------------
man_pages = [
    ('index', 'glitter_python_driver',
     'Glitter Python Driver Documentation',
     ['ted'], 1)
]





# Copyright GlitterClient GmbH and GlitterClient contributors
# SPDX-License-Identifier: (Apache-2.0 AND CC-BY-4.0)
# Code is Apache-2.0 and docs are CC-BY-4.0
import random
from abc import ABCMeta, abstractmethod
from datetime import datetime


class AbstractPicker(metaclass=ABCMeta):
    """Abstract class for picker classes that pick connections from a pool."""

    @abstractmethod
    def pick(self, connections):
        """Picks a :class:`~GlitterClient_driver.connection.Connection`
        instance from the given list of
        :class:`~GlitterClient_driver.connection.Connection` instances.

        Args:
            connections (List): List of
                :class:`~GlitterClient_driver.connection.Connection` instances.

        """
        pass    # pragma: no cover


class RoundRobinPicker(AbstractPicker):
    """Picks a :class:`~GlitterClient_driver.connection.Connection`
       instance from a list of connections.

    """

    def pick(self, connections):
        """Picks a connection with the earliest backoff time.

           As a result, the first connection is picked
           for as long as it has no backoff time.
           Otherwise, the connections are tried in a round robin fashion.

        Args:
            connections (:obj:list): List of
                :class:`~GlitterClient_driver.connection.Connection` instances.

        """
        if len(connections) == 1:
            return connections[0]

        def key(conn):
            ms = random.randrange(0, 1000)
            return (datetime(1, 1, 1, 0, 0, 0, ms)
                    if conn.backoff_time is None
                    else conn.backoff_time)
        return min(*connections, key=key)


class Pool:
    """Pool of connections."""

    def __init__(self, connections, picker_class=RoundRobinPicker):
        """Initializes a :class:`~GlitterClient_driver.pool.Pool` instance.

        Args:
            connections (list): List of
                :class:`~GlitterClient_driver.connection.Connection` instances.

        """
        self.connections = connections
        self.picker = picker_class()

    def get_connection(self):
        """Gets a :class:`~GlitterClient_driver.connection.Connection`
        instance from the pool.

        Returns:
            A :class:`~GlitterClient_driver.connection.Connection` instance.

        """
        return self.picker.pick(self.connections)



# Copyright GlitterClient GmbH and GlitterClient contributors
# SPDX-License-Identifier: (Apache-2.0 AND CC-BY-4.0)
# Code is Apache-2.0 and docs are CC-BY-4.0

# from .driver import GlitterClient   # noqa
from .driver import GlitterClient   # noqa


__author__ = 'ted'
__email__ = 'ted@glitterprotocol.io'
__version__ = '0.1'
__release__ = '0.1.1'



# Copyright GlitterClient GmbH and GlitterClient contributors
# SPDX-License-Identifier: (Apache-2.0 AND CC-BY-4.0)
# Code is Apache-2.0 and docs are CC-BY-4.0

from .transport import Transport
from .utils import normalize_nodes

class GlitterClient:
    """A :class: `~driver.GlitterClient` is python client  for glitter. It can be connect, create schema, put docs and search .

    """

    def __init__(self, *nodes, headers=None, transport_class=Transport, timeout=10):
        """Initialize a :class:`~driver.GlitterClient` driver instance.

        Args:
            *nodes:(list of (str or dict)): Glitter nodes to connect to.
            headers (dict): Optional headers that will be passed with each request
            transport_class: Optional transport class to use.
            timeout (int): Optional timeout in seconds that will be passed to each request.
        """
        self._headers = headers
        self._nodes = normalize_nodes(*nodes, headers=headers)
        self._transport = transport_class(*self._nodes, timeout=timeout)
        self._db = DataBase(self)
        self._chain = Chain(self)
        self._admin = Admin(self)
        self.api_prefix = '/v1'

    @property
    def nodes(self):
        """:obj:`tuple` of :obj:`str`:
         URLs of connected nodes.
        """
        return self._nodes

    @property
    def transport(self):
        """:class:`~driver.Transport`: Object responsible for forwarding requests to a :class:`~driver.Connection` instance (node).
        """
        return self._transport

    @property
    def chain(self):
        """:class:`~driver.Chain`: query block or transaction info.
        """
        return self._chain

    @property
    def admin(self):
        """:class:`~driver.Admin`: Exposes functionalities of the ``'/admin'`` endpoint.
        """
        return self._admin

    @property
    def db(self):
        """:class:`~driver.DataBase`: put or search doc from glitter.
        """
        return self._db


class NamespacedDriver:
    """Base class for creating endpoints (namespaced objects) that can be added
    under the :class:`~driver.GlitterClient` driver.
    """

    PATH = '/'

    def __init__(self, driver):
        """Initializes an instance of
        :class:`~GlitterClient_driver.driver.NamespacedDriver` with the given
        driver instance.

        Args:
            driver (GlitterClient): Instance of
                :class:`~GlitterClient_driver.driver.GlitterClient`.
        """
        self.driver = driver

    @property
    def transport(self):
        return self.driver.transport

    @property
    def api_prefix(self):
        return self.driver.api_prefix

    @property
    def path(self):
        return self.api_prefix + self.PATH


class DataBase(NamespacedDriver):
    """Exposes the data of glitter db.
    """

    def create_schema(self, schema_name, fields):
        """

        Args:
            - schema_name (str): the name of schema.
            - fields (:obj:`list` of :obj:`dic`): list of schema fields.

        Returns:
            - :obj:`dic`: request result.
        """
        path = '/create_schema'
        schema_type = "record"
        body = {
            "schema_name": schema_name,
            "data": {
                "type": schema_type,
                "name": schema_name,
                "fields": fields
            }
        }

        return self.transport.forward_request(
            method='POST',
            path=self.api_prefix + path,
            json=body,
        )

    def list_schema(self):
        """

        Returns:
            - :obj:`dic`: list all schema.
        """
        path = '/list_schema'

        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
        )

    def get_schema(self, schema_name):
        """

        Args:
            - schema_name(str): the name of schema.
        Returns:
            - :obj:`dic`: result with document schema.
        """
        path = '/get_schema'

        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
            params={"schema_name": schema_name}
        )

    def app_status(self):
        """

        Returns:
            - :obj:`dic`: result with document count.
        """

        path = '/app_status'

        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
        )

    def put_doc(self, schema_name, doc_data):
        """Put document to glitter.

        Args:
            - schema_name(str): the name of schema.
            - doc_data(:obj:`dic`): doc content.

        Returns:
            - :obj:`dic`: result with code,msg,tx.
        """
        path = '/put_doc'
        body = {
            "schema_name": schema_name,
            "doc_data": doc_data,
        }
        return self.transport.forward_request(
            method='POST',
            path=self.api_prefix + path,
            json=body,
        )

    def get_doc(self, schema_name, primary_key):
        """Get documents from glitter by doc ids.

        Args:
            schema_name(str): the name of schema.
            primary_key: main key of document,must be uniq.

        Returns:
            :obj:`dic`: result with the document struct.
        """
        path = '/get_docs'
        return self.transport.forward_request(
            method='POST',
            path=self.api_prefix + path,
            json={"schema_name": schema_name, "doc_ids": [primary_key]},
        )

    def get_docs(self, schema_name, primary_key):
        """Get documents from glitter by doc ids.

        Args:
            schema_name(str): the name of schema.
            primary_key(list): main key of documents,must be uniq.

        Returns:
            :obj:`dic`: result with the document struct.
        """
        path = '/get_docs'
        return self.transport.forward_request(
            method='POST',
            path=self.api_prefix + path,
            json={"schema_name": schema_name, "doc_ids": primary_key},
        )

    def search(self, index, query_word, query_field=[], filters=[], aggs_field=[], order_by="", limit=10, page=1):
        """ search from glitter,with more args.

        Args:
            index(str): index name.
            query_word(str): query word, only applies to  query_field.
            query_field(:obj: `list` of str): query field must be indexed in schema.
            filters(:obj:`list` of :obj:`dic`): filter condition.
            aggs_field(:obj: `list` of str): aggregate field ,which is define in schema
            order_by(str): order field
            limit(int): limit
            page(int): page number,begin from 1

        Returns:
            :obj:`dic`: the documents match query words.
        """

        path = '/search'
        return self.transport.forward_request(
            method='POST',
            path=self.api_prefix + path,
            json={"index": index, "query": query_word, "filters": filters, "query_field": query_field,
                  "aggs_field": aggs_field, "order_by": order_by, "limit": limit, "page": page},
        )


class Chain(NamespacedDriver):
    PATH = '/chain/'

    def status(self):
        """ Get Tendermint status including node info, pubkey, latest block hash, app hash, block height, current max peer height, and time.

        Returns:
            :obj:`json`:Details of the HTTP API provided by the tendermint server.
        """
        path = "/chain/status"
        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
        )

    def tx_search(self, query, page=1, per_page=30, order_by="desc", prove=True):
        """ Search for transactions their results

        Args:
            query(str): query words. (e.g: ``tx.height=1000, tx.hash='xxx', update_doc.token='test_token'``)
            page(int): page number. Defaults to ``1``.
            per_page(int): number of entries per page (max: ``100``).Defaults to ``30``.
            order_by(str): Order in which transactions are sorted (``asc`` or ``desc``), by height & index. If empty, default sorting will be still applied.
            prove(bool): Include proofs of the transactions inclusion in the block. Defaults to ``True``.

        Returns:
            :obj"`json`: transaction info
        """

        path = "/chain/tx_search"
        query = '"{}"'.format(query)
        order_by = '"{}"'.format(order_by)
        prove = "true" if prove else "false"

        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
            params={'query': query, 'page': page, 'per_page': per_page, 'order_by': order_by, 'prove': prove},
        )

    def block_search(self, query, page=1, per_page=30, order_by="desc"):
        """Search for blocks by BeginBlock and EndBlock events

        Args:
            query(str): query condition. (e.g: ``block.height > 1000 AND valset.changed > 0``)
            page(int): page number
            per_page(int): number of entries per page (max: 100)
            order_by(str): order in which blocks are sorted ("asc" or "desc"), by height. If empty, default sorting will be still applied.
        Returns:
            :obj:`json`: block info

        """
        path = "/chain/block_search"
        query = '"{}"'.format(query)
        order_by = '"{}"'.format(order_by)

        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
            params={'query': query, 'page': page, 'per_page': per_page, 'order_by': order_by},
        )

    def block(self, height=None):
        """ Get block at a specified height

        Args:
            height(int): height

        Returns:
            :obj:`json`:height to return. If no height is provided, it will fetch the latest block.

        """
        path = "/chain/block"
        params = {}
        if height is not None:
            params["height"] = height
        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
            params=params,
        )

    def health(self):
        """ Get node health.

        Returns:
            Details of the HTTP API provided by the tendermint server, empty result (200 OK) on success, no response - in case of an error.
        """
        path = "/chain/health"
        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
        )

    def net_info(self):
        """ Get network info.

        Returns:
            Details of the HTTP API provided by the tendermint server.

        """
        path = "/chain/net_info"
        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
        )

    def blockchain(self, min_height=1, max_height=20):
        """ Get block headers for minHeight <= height maxHeight.
        If maxHeight does not yet exist, blocks up to the current height will be returned. If minHeight does not exist (due to pruning), earliest existing height will be used.
        At most 20 items will be returned. Block headers are returned in descending order (highest first).

        Args:
            min_height(int): Minimum block height to return
            max_height(bool): Maximum block height to return

        Returns:
            Block headers, returned in descending order (highest first).
        """

        path = "/chain/blockchain"
        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
            params={"minHeight": min_height, "maxHeight": max_height}
        )

    def header(self, height=1):
        """ Retrieve the block header corresponding to a specified height.

        Args:
            height(int): height to return. If no height is provided, it will fetch the latest height.

        Returns:
            Header information.

        """
        path = "/chain/header"
        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
            params={"height": height},
        )

    def header_by_hash(self, header_hash):
        """ Retrieve the block header corresponding to a block hash.

        Args:
            header_hash(str): header hash
        Returns:
        """
        path = "/chain/header_by_hash"
        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
            params={"hash": header_hash},
        )

    def block_by_hash(self, *, header_hash):
        """ Get block by hash

        Args:
            header_hash(str): block hash. example: "0xD70952032620CC4E2737EB8AC379806359D8E0B17B0488F627997A0B043ABDED"

        Returns:

        """
        path = "/chain/header_by_hash"
        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
            params={"hash": header_hash},
        )


class Admin(NamespacedDriver):
    """Exposes functionality of the ``'/admin'`` endpoint.
    """

    def update_validator(self, pub_key, power=0, headers=None):
        """ update validator set

        Args:
            pub_key (str): public key
            power (int): power
            headers (dict): http header

        Returns:
            :obj:`dic`:

        """
        path = "/admin/update_validator"
        return self.transport.forward_request(
            method='POST',
            path=self.api_prefix + path,
            params={'pub_key': pub_key, 'power': power},
            headers=headers
        )

    def validators(self, height=None, page=1, per_page=100):
        """ Get validator set at a specified height

        Args:
            height (str): height to return. If no height is provided, it will fetch validator set which corresponds to the latest block.
            page (int): Page number (1-based)
            per_page (int): Number of entries per page (max: 100)

        Returns:
            :obj:`json`: Validators. Validators are sorted first by voting power (descending), then by address (ascending).

        """
        path = "/chain/validators"
        params = {"page": page, "per_page": per_page}
        if height is not None:
            params["height"] = height

        return self.transport.forward_request(
            method='GET',
            path=self.api_prefix + path,
            params=params,
        )



# Copyright GlitterClient GmbH and GlitterClient contributors
# SPDX-License-Identifier: (Apache-2.0 AND CC-BY-4.0)
# Code is Apache-2.0 and docs are CC-BY-4.0

import time

from collections import namedtuple
from datetime import datetime, timedelta

from requests import Session
from requests.exceptions import ConnectionError
from requests.exceptions import RequestException

from .exceptions import HTTP_EXCEPTIONS, TransportError,TimeoutError


BACKOFF_DELAY = 0.5  # seconds

HttpResponse = namedtuple('HttpResponse', ('status_code', 'headers', 'data'))


class Connection:
    """A Connection object to make HTTP requests to a particular node."""

    def __init__(self, *, node_url, headers=None):
        """Initializes a :class:`~GlitterClient_driver.connection.Connection`
        instance.

        Args:
            node_url (str):  Url of the node to connect to.
            headers (dict): Optional headers to send with each request.

        """
        self.node_url = node_url
        self.session = Session()
        if headers:
            self.session.headers.update(headers)

        self._retries = 0
        self.backoff_time = None

    def request(self, method, *, path=None, json=None,
                params=None, headers=None, timeout=None,
                backoff_cap=None, **kwargs):
        """Performs an HTTP request with the given parameters.

           Implements exponential backoff.

           If `ConnectionError` occurs, a timestamp equal to now +
           the default delay (`BACKOFF_DELAY`) is assigned to the object.
           The timestamp is in UTC. Next time the function is called, it either
           waits till the timestamp is passed or raises `TimeoutError`.

           If `ConnectionError` occurs two or more times in a row,
           the retry count is incremented and the new timestamp is calculated
           as now + the default delay multiplied by two to the power of the
           number of retries.

           If a request is successful, the backoff timestamp is removed,
           the retry count is back to zero.

        Args:
            method (str): HTTP method (e.g.: ``'GET'``).
            path (str): API endpoint path (e.g.: ``'/transactions'``).
            json (dict): JSON data to send along with the request.
            params (dict): Dictionary of URL (query) parameters.
            headers (dict): Optional headers to pass to the request.
            timeout (int): Optional timeout in seconds.
            backoff_cap (int): The maximal allowed backoff delay in seconds
                               to be assigned to a node.
            kwargs: Optional keyword arguments.

        """
        backoff_timedelta = self.get_backoff_timedelta()

        if timeout is not None and timeout < backoff_timedelta:
            raise TimeoutError

        if backoff_timedelta > 0:
            time.sleep(backoff_timedelta)

        connExc = None
        timeout = timeout if timeout is None else timeout - backoff_timedelta
        try:
            response = self._request(
                method=method,
                timeout=timeout,
                url=self.node_url + path if path else self.node_url,
                json=json,
                params=params,
                headers=headers,
                **kwargs,
            )
        except RequestException as err:
            connExc = err
            raise err
        except TransportError as err:
            connExc = err
            raise err
        finally:
            self.update_backoff_time(success=connExc is None,
                                     backoff_cap=backoff_cap)
        return response

    def get_backoff_timedelta(self):
        if self.backoff_time is None:
            return 0

        return (self.backoff_time - datetime.utcnow()).total_seconds()

    def update_backoff_time(self, success, backoff_cap=None):
        if success:
            self._retries = 0
            self.backoff_time = None
        else:
            utcnow = datetime.utcnow()
            backoff_delta = BACKOFF_DELAY * 2 ** self._retries
            if backoff_cap is not None:
                backoff_delta = min(backoff_delta, backoff_cap)
            self.backoff_time = utcnow + timedelta(seconds=backoff_delta)
            self._retries += 1

    def _request(self, **kwargs):
        response = self.session.request(**kwargs)
        text = response.text
        try:
            json = response.json()
        except ValueError:
            json = None
        if not (200 <= response.status_code < 300):
            exc_cls = HTTP_EXCEPTIONS.get(response.status_code, TransportError)
            raise exc_cls(response.status_code, text, json, kwargs['url'])
        data = json if json is not None else text
        return HttpResponse(response.status_code, response.headers, data)



# Copyright GlitterClient GmbH and GlitterClient contributors
# SPDX-License-Identifier: (Apache-2.0 AND CC-BY-4.0)
# Code is Apache-2.0 and docs are CC-BY-4.0

"""Exceptions used by :mod:`GlitterClient_driver`."""


class GlitterClientException(Exception):
    """Base exception for all GlitterClient exceptions."""


class KeypairNotFoundException(GlitterClientException):
    """Raised if an operation cannot proceed because the keypair
    was not given.
    """


class InvalidPrivateKey(GlitterClientException):
    """Raised if a private key is invalid. E.g.: :obj:`None`."""


class InvalidPublicKey(GlitterClientException):
    """Raised if a public key is invalid. E.g.: :obj:`None`."""


class MissingPrivateKeyError(GlitterClientException):
    """Raised if a private key is missing."""


class TimeoutError(GlitterClientException):
    """Raised if the request algorithm times out."""

    @property
    def connection_errors(self):
        """Returns connection errors occurred before timeout expired."""
        return self.args[0]


class TransportError(GlitterClientException):
    """Base exception for transport related errors.

    This is mainly for cases where the status code denotes an HTTP error, and
    for cases in which there was a connection error.

    """

    @property
    def status_code(self):
        return self.args[0]

    @property
    def error(self):
        return self.args[1]

    @property
    def info(self):
        return self.args[2]

    @property
    def url(self):
        return self.args[3]


class ConnectionError(TransportError):
    """Exception for errors occurring when connecting, and/or making a request
    to GlitterClient.

    """


class BadRequest(TransportError):
    """Exception for HTTP 400 errors."""


class NotFoundError(TransportError):
    """Exception for HTTP 404 errors."""


class ServiceUnavailable(TransportError):
    """Exception for HTTP 503 errors."""


class GatewayTimeout(TransportError):
    """Exception for HTTP 503 errors."""


HTTP_EXCEPTIONS = {
    400: BadRequest,
    404: NotFoundError,
    503: ServiceUnavailable,
    504: GatewayTimeout,
}



# Copyright 2022-present glitter, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the glitter driver module."""

import unittest
from glitter_sdk import GlitterClient


class GlitterClientUnitTest(unittest.TestCase):
    glitter_client: GlitterClient
    schema_name = "sample"

    @classmethod
    def setUpClass(cls):
        url = 'http://sg1.testnet.glitter.link:26659'
        cls.glitter_client = GlitterClient()

    def test_create_schema(self):
        fields = [
            {
                "name": "doi",
                "type": "string",
                "primary": "true",
                "index": {
                    "type": "keyword"
                }
            },
            {
                "name": "title",
                "type": "string",
                "index": {
                    "type": "text"
                }
            },
            {
                "name": "ipfs_cid",
                "type": "string",
                "index": {
                    "index": "false"
                }
            }
        ]
        res = self.glitter_client.db.create_schema(self.schema_name, fields)
        # self.assertEqual(res['code'], 0)
        print(res)

    def test_get_schema(self):
        res = self.glitter_client.db.get_schema(self.schema_name)
        print(res)

    def test_list_schema(self):
        res = self.glitter_client.db.list_schema()
        self.assertEqual(res['code'], 0)
        self.assertIsNotNone(res['data'].get('demo'))
        print(res)

    def test_put_doc(self):
        demo_doc = {
            "doi": "10.1002/(sci)1099-1697(199803/04)7:2<65::aid-jsc357>3.0.c",
            "title": "British Steel Corporation: probably the biggest turnaround story in UK industrial history",
            "ipfs_cid": "bafybeibxvp6bawmr4u24vuza2vyretip4n7sfvivg7hdbyolxrvbodwlte"
        }

        res = self.glitter_client.db.put_doc(self.schema_name, demo_doc)
        self.assertEqual(res['code'], 0)
        self.assertIsNotNone(res['tx'])
        print(res)

    def test_get_docs(self):
        primary_key = "10.1002/(sci)1099-1697(199803/04)7:2<65::aid-jsc357>3.0.c"
        res = self.glitter_client.db.get_docs(self.schema_name, [primary_key])
        print(res)
        # self.assertEqual(res["code"], 0)
        # self.assertGreaterEqual(res["data"]["Total"], 1)
        # self.assertEqual(res["data"]["hits"][primary_key]["doc_id"], primary_key)
        # print(res)

    def test_search(self):
        query_word = "British Steel Corporation"
        query_field = ["doi", "title"]
        res = self.glitter_client.db.search("xx", query_word, query_field)
        print(res)
        # self.assertEqual(res["code"], 0)
        # self.assertGreaterEqual(res["data"]["meta"]["page"]["size"], 1)

    def test_search2(self):
        ''' First, put  documents:
            {
                "doi": "doi_1",
                "title": "British Steel Corporation: probably the biggest turnaround story in UK industrial history",
                "ipfs_cid": "bafybeibxvp6bawmr4u24vuza2vyretip4n7sfvivg7hdbyolxrvbodwlte",
                "publish_year": 1992
            }
            {
                "doi": "doi_2",
                "title": "British Steel Corporation: probably the biggest turnaround story in UK industrial history",
                "ipfs_cid": "bafybeibxvp6bawmr4u24vuza2vyretip4n7sfvivg7hdbyolxrvbodwlte",
                "publish_year": 2022
            }

        '''
        query_field = ["doi", "title"]

        # range query, 1990 <= publish_year <= 2000
        # range_conds = [{"type": "range", "field": "publish_year", "from": 1990, "to": 2000}]
        # res = self.glitter_client.db.search(self.schema_name, "British Steel Corporation",query_field, range_conds )

        term_conds = [{"type": "term", "field": "publish_year", "value": 1992}]
        res = self.glitter_client.db.search(self.schema_name, "British Steel Corporation",query_field, term_conds)
        print(res)
        # self.assertEqual(res["code"], 0)
        # self.assertGreaterEqual(res["data"]["meta"]["page"]["size"], 1)

    def test_app_status(self):
        res = self.glitter_client.db.app_status()
        self.assertEqual(res['code'], 0)
        print(res)

    def test_block(self):
        res = self.glitter_client.chain.block()
        print(res)

    def test_block(self):
        res = self.glitter_client.chain.block_search(query="block.height = 17835")
        print(res)

    def test_others(self):
        # res = self.glitter_client.db.search("sample", "Content Indexing Network")
        # print(res)
        for i in range(1, 100):
            try:
                res = self.glitter_client.db.get_schema(self.schema_name)
                print(res)
            except Exception as e:
                print(e)
                continue


if __name__ == '__main__':
    unittest.main()



# Copyright GlitterClient GmbH and GlitterClient contributors
# SPDX-License-Identifier: (Apache-2.0 AND CC-BY-4.0)
# Code is Apache-2.0 and docs are CC-BY-4.0

from time import time

from requests.exceptions import ConnectionError

from .connection import Connection
from .exceptions import TimeoutError
from .pool import Pool


NO_TIMEOUT_BACKOFF_CAP = 10  # seconds


class Transport:
    """Transport class.

    """

    def __init__(self, *nodes, timeout=None):
        """Initializes an instance of
        :class:`~GlitterClient_driver.transport.Transport`.

        Args:
            nodes: each node is a dictionary with the keys `endpoint` and
                   `headers`
            timeout (int): Optional timeout in seconds.

        """
        self.nodes = nodes
        self.timeout = timeout
        self.connection_pool = Pool([Connection(node_url=node['endpoint'],
                                                headers=node['headers'])
                                     for node in nodes])

    def forward_request(self, method, path=None,
                        json=None, params=None, headers=None):
        """Makes HTTP requests to the configured nodes.

           Retries connection errors
           (e.g. DNS failures, refused connection, etc).
           A user may choose to retry other errors
           by catching the corresponding
           exceptions and retrying `forward_request`.

           Exponential backoff is implemented individually for each node.
           Backoff delays are expressed as timestamps stored on the object and
           they are not reset in between multiple function calls.

           Times out when `self.timeout` is expired, if not `None`.

        Args:
            method (str): HTTP method name (e.g.: ``'GET'``).
            path (str): Path to be appended to the base url of a node. E.g.:
                ``'/transactions'``).
            json (dict): Payload to be sent with the HTTP request.
            params (dict)): Dictionary of URL (query) parameters.
            headers (dict): Optional headers to pass to the request.

        Returns:
            dict: Result of :meth:`requests.models.Response.json`

        """
        error_trace = []
        timeout = self.timeout
        backoff_cap = NO_TIMEOUT_BACKOFF_CAP if timeout is None \
            else timeout / 2
        while timeout is None or timeout > 0:
            connection = self.connection_pool.get_connection()

            start = time()
            try:
                response = connection.request(
                    method=method,
                    path=path,
                    params=params,
                    json=json,
                    headers=headers,
                    timeout=timeout,
                    backoff_cap=backoff_cap,
                )
            except ConnectionError as err:
                error_trace.append(err)
                continue
            else:
                return response.data
            finally:
                elapsed = time() - start
                if timeout is not None:
                    timeout -= elapsed

        raise TimeoutError(error_trace)



#!/usr/bin/env python
# coding: utf-8

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('./docs/install.rst') as install_file:
    install = install_file.read()

with open('./docs/tutorial.rst') as tutorial_file:
    tutorial = tutorial_file.read()

with open('CHANGELOG.rst') as changelog_file:
    changelog = changelog_file.read()

tests_require = [
    'pytest>=7.0.1',
]

dev_require = [
    'ipdb',
    'ipython',
    'pre-commit'
]

docs_require = [
    'Sphinx~=4.0',
    'sphinx-autobuild',
    'sphinxcontrib-autorun',
    'sphinxcontrib-napoleon>=0.4.4',
    'sphinx_rtd_theme',
    'sphinxcontrib-httpdomain',
    'sphinx-tabs',
]
setup(
    name='glitter_sdk',
    version='0.1.1',
    author='Glitter Protocol',
    author_email='ted@glitterprotocol.io',
    url='https://docs.glitterprotocol.io/',
    description=u'Glitter Protocol is a blockchain based database and index engine for developing and hosting web3 applications in decentralized storage networks.',
    long_description=readme + '\n\n' + changelog,
    python_requires='>=3.5',
    packages=find_packages(),
    install_requires=['requests'],
    entry_points={},
    test_suite='tests',
    extras_require={
        'test': tests_require,
        'dev': dev_require + tests_require + docs_require,
        'docs': docs_require,
    },
)
